import urllib.parse

xsd = "http://www.w3.org/2001/XMLSchema#"


class Service(object):

    def __init__(self, endpoint, triples, limit=-1, filter_nested=None):
        self.endpoint = endpoint[1:-1] if endpoint[0] == "<" else endpoint
        self.triples = triples
        self.filters = []
        self.filter_nested = filter_nested if filter_nested is not None else []  # TODO: this is used to store the filters from NestedLoop operators
        self.limit = limit  # TODO: This arg was added in order to integrate contactSource with incremental calls (16/12/2013)
        #self.filters_vars = set(filter_vars)

    def include_filter(self, f):
        self.filters.append(f)

    def __add__(self, other):
        self.triples.extend(other.triples)
        self.filters.extend(other.filters)
        self.filter_nested.extend(other.filter_nested)
        if other.limit > self.limit:
            self.limit = other.limit

    def __repr__(self):
        if isinstance(self.triples, list):
            triples_str = " . ".join(map(str, self.triples))
        else:
            triples_str = str(self.triples)
        filters_str = " . ".join(map(str, self.filters)) + " \n".join(map(str, self.filter_nested))
        return " { SERVICE <" + self.endpoint + "> { " + triples_str + filters_str + " }   \n }"

    def allTriplesGeneral(self):
        a = True
        if isinstance(self.triples, list):
            for t in self.triples:
                a = a and t.allTriplesGeneral()
        else:
            a = self.triples.allTriplesGeneral()
        return a

    def allTriplesLowSelectivity(self):
        a = True
        if isinstance(self.triples, list):
            for t in self.triples:
                a = a and t.allTriplesLowSelectivity()
        else:
            a = self.triples.allTriplesLowSelectivity()
        a = a or (self.filters != [])
        return a

    def instantiate(self, d):
        if isinstance(self.triples, list):
             new_triples = [t.instantiate(d) for t in self.triples]
        else:
             new_triples = self.triples.instantiate(d)
        return Service("<" + self.endpoint + ">", new_triples, self.limit)

    def instantiateFilter(self, d, filter_str):
        new_filters = []
        new_filters.extend(self.filter_nested)
        new_filters.append(filter_str)
        #new_filters_vars = self.filters_vars | set(d)

        return Service("<" + self.endpoint + ">", self.triples, self.limit, new_filters)

    def getTriples(self):
        if isinstance(self.triples, list):
            triples_str = " . ".join(map(str, self.triples))
        else:
            triples_str = str(self.triples)
        return triples_str + " . ".join(map(str, self.filters)) + " . ".join(map(str, self.filter_nested))

    def show(self, x):
        def pp(t):
            return t.show(x + "    ")
        if isinstance(self.triples, list):
            triples_str = " . \n".join(map(pp, self.triples))
        else:
            triples_str = self.triples.show(x+"    ")
        filters_str = " . \n".join(map(pp, self.filters)) + "  \n".join(map(pp, self.filter_nested))

        return x + "SERVICE <" + self.endpoint + "> { \n" + triples_str + filters_str + "\n" + x + "}"

    def show2(self, x):
        def pp(t):
            return t.show2(x + "    ")
        if isinstance(self.triples, list):
            triples_str = " . \n".join(map(pp, self.triples))
        else:
            triples_str = self.triples.show2(x+"    ")
        filters_str = " . \n".join(map(pp, self.filters)) + "  \n".join(map(pp, self.filter_nested))
        return triples_str + filters_str

    def getVars(self):
        if isinstance(self.triples, list):
            l = []
            for t in self.triples:
                l = l + t.getVars()
        else:
            l = self.triples.getVars()
        return l

    def getConsts(self):
        if isinstance(self.triples, list):
            c = []
            for t in self.triples:
                c = c + t.getConsts()
        else:
            c = self.triples.getConsts()

        return c

    def getPredVars(self):
        if isinstance(self.triples, list):
            l = []
            for t in self.triples:
                l = l + t.getPredVars()
        else:
            l = self.triples.getPredVars()
        return l

    def places(self):
        p = 0
        if isinstance(self.triples, list):
            for t in self.triples:
                p = p + t.places()
        else:
            p = self.triples.places()
        return p

    def constantNumber(self):
        p = 0
        if isinstance(self.triples, list):
            for t in self.triples:
                p = p + t.constantNumber()
        else:
            p = self.triples.constantNumber()
        return p

    def constantPercentage(self):
        if self.places() == 0:
            return 0
        return self.constantNumber()/self.places()

    def setGeneral(self, ps, genPred):
        if isinstance(self.triples, list):
            for t in self.triples:
                t.setGeneral(ps, genPred)
        else:
            self.triples.setGeneral(ps, genPred)


class Query(object):

    def __init__(self, prefs, args, body, distinct, values, order_by=None, limit=-1, offset=-1, filter_nested=''):
        self.prefs = prefs
        self.args = args
        self.body = body
        self.distinct = distinct
        self.values = values
        self.join_vars = self.getJoinVars()
        self.order_by = order_by if order_by is not None else []
        self.limit = limit
        self.offset = offset
        self.filter_nested = filter_nested
        genPred = [] #readGeneralPredicates('ontario/common/parser/generalPredicates')
        self.body.setGeneral(getPrefs(self.prefs), genPred)

    def __repr__(self):
        body_str = str(self.body)
        args_str = " ".join(map(str, self.args))
        if self.args == []:
            args_str = "*"
        if self.distinct:
            d = "DISTINCT "
        else:
            d = ""
        values_str = str(self.values)
        return self.getPrefixes() + "SELECT " + d + args_str + "\nWHERE {" + body_str + "\n" + values_str + "\n" + self.filter_nested + "\n}"

    def variables(self):
        if not self.args:
            return [var.replace('?', '') for var in set(self.body.getVars())]
        else:
            return [str(arg).replace('?', '') for arg in self.args]

    def instantiate(self, d):
        new_args = []
        # for a in self.args:
        #     an = string.lstrip(string.lstrip(self.subject.name, "?"), "$")
        #     if not (an in d):
        #         new_args.append(a)
        return Query(self.prefs, new_args, self.body.instantiate(d), self.distinct, self.values)

    def instantiateFilter(self, d, filter_str):
        new_args = []
        # for a in self.args:
        #     an = string.lstrip(string.lstrip(self.subject.name, "?"), "$")
        #     if not (an in d):
        #         new_args.append(a)
        return Query(self.prefs, new_args, self.body, self.distinct, self.values, self.filter_nested + ' ' + filter_str)

    def places(self):
        return self.body.places()

    def constantNumber(self):
        return self.body.constantNumber()

    def constantPercentage(self):
        return self.constantNumber()/self.places()

    def show(self):
        body_str = self.body.show(" ")
        args_str = " ".join(map(str, self.args))
        if self.args == []:
            args_str = "*"
        if self.distinct:
            d = "DISTINCT "
        else:
            d = ""
        values_str = str(self.values)
        return self.getPrefixes() + "SELECT " + d + args_str + "\nWHERE {" + body_str + "\n" + values_str  + "\n" + self.filter_nested + "\n}"

    def show2(self):
        body_str = self.body.show2(" ")
        args_str = " ".join(map(str, self.args))
        if self.args == []:
            args_str = "*"
        if self.distinct:
            d = "DISTINCT "
        else:
            d = ""
        values_str = str(self.values)
        return self.getPrefixes() + "SELECT " + d + args_str + "\nWHERE {" + body_str + "\n" + values_str + "\n" + self.filter_nested + "\n}"

    def getPrefixes(self):
        r = ""
        for e in self.prefs:
            i = e.find(":")
            e = e[:i] + ": " + e[i+1:]
            r = r + "\nprefix " + e
        if not r == "":
            r = r + "\n"
        return r

    def getJoinVars(self):
        join_vars = getJoinVarsUnionBlock(self.body)
        join_vars = [v for v in join_vars if join_vars.count(v) > 1]

        return set(join_vars)

    def getFilterVars(self):
        filtervars = getFilterVarsUnionBlock(self.body)
        return set(filtervars)

    def getJoinVars2(self):
        join_vars = []

        for s in self.body:
            for t in s.triples:
                if not t.subject.constant:
                    join_vars.append(t.subject.name)
                if not t.theobject.constant:
                    join_vars.append(t.theobject.name)

        join_vars = [v for v in join_vars if join_vars.count(v) > 1]

        return set(join_vars)

    def getTreeRepresentation(self):
        l0 = self.body
        while len(l0) > 1:
            l1 = []
            while len(l0) > 1:
                x = l0.pop()
                y = l0.pop()
                l1.append((x, y))
            if len(l0) == 1:
                l1.append(l0.pop())
            l0 = l1
        if len(l0) == 1:
            return aux(l0[0], "", " xxx ")
        else:
            return " "


def getFilterVarsUnionBlock(ub):
    filter_vars = []
    if len(ub.filters) > 0:
        for f in ub.filters:
            filter_vars.extend(f.getVars())

    for jb in ub.triples:
        filter_vars.extend(getFilterVarsJoinBlock(jb))

    return filter_vars


def getFilterVarsJoinBlock(jb):
    join_vars = []
    if len(jb.filters) > 0:
        for f in jb.filters:
            join_vars.extend(f.getVars())

    if isinstance(jb, JoinBlock) and isinstance(jb.triples, list):

        for bgp in jb.triples:
            if isinstance(bgp, Service):
                join_vars.extend(getFilterVarsUnionBlock(bgp.triples))
            elif isinstance(bgp, Optional):
                join_vars.extend(getFilterVarsUnionBlock(bgp.bgg))
            elif isinstance(bgp, UnionBlock):
                join_vars.extend(getFilterVarsUnionBlock(bgp))
            elif isinstance(bgp, JoinBlock):
                join_vars.extend(getFilterVarsJoinBlock(bgp))
    return join_vars


def getJoinVarsUnionBlock(ub):
    join_vars = []

    for jb in ub.triples:
        join_vars.extend(getJoinVarsJoinBlock(jb))

    return join_vars


def getJoinVarsJoinBlock(jb):
    join_vars = []

    for bgp in jb.triples:

        if isinstance(bgp, Triple):
            if not bgp.subject.constant:
                join_vars.append(bgp.subject.name)
            if not bgp.theobject.constant:
                join_vars.append(bgp.theobject.name)
        elif isinstance(bgp, Service):
            join_vars.extend(getJoinVarsUnionBlock(bgp.triples))
        elif isinstance(bgp, Optional):
            join_vars.extend(getJoinVarsUnionBlock(bgp.bgg))
        elif isinstance(bgp, UnionBlock):
            join_vars.extend(getJoinVarsUnionBlock(bgp))
        elif isinstance(bgp, JoinBlock):
            join_vars.extend(getJoinVarsJoinBlock(bgp))
    return join_vars


def aux(e, x, op):
    def pp(t):
        return t.show(x+"  ")
    if type(e) == tuple:
        (f, s) = e
        r = ""
        if f:
            r = x + "{\n" + aux(f, x + "  ", op) + "\n" + x + "}\n"
        if f and s:
            r = r + x + op + "\n"
        if s:
            r = r + x + "{\n" + aux(s, x + "  ", op) + "\n" + x + "}"
        return r
    elif type(e) == list:
        return (x + " . \n").join(map(pp, e))
    elif e:
        return e.show(x + "  ")

    return ""


def aux2(e, x, op):
    def pp(t):
        return t.show2(x + "  ")
    if type(e) == tuple:
        (f, s) = e
        r = ""
        if f:
            r = x + "{\n" + aux2(f, x + "  ", op) + "\n" + x + "}\n"
        if f and s:
            r = r + x + op + "\n"
        if s:
            r = r + x + "{\n" + aux2(s, x + "  ", op) + "\n" + x + "}"
        return r
    elif type(e) == list:
        return (x + " . \n").join(map(pp, e))
    elif e:
        return e.show2(x + "  ")
    return ""


class UnionBlock(object):

    def __init__(self, triples,filters=[]):
        self.triples = triples
        self.filters = filters

    def __repr__(self):
        return self.show(" ")

    def show(self, w):

        n = nest(self.triples)

        if n:
            return aux(n, w, " UNION ") + " ".join(map(str, self.filters))
        else:
            return " "

    def setGeneral(self, ps, genPred):
        if isinstance(self.triples, list):
            for t in self.triples:
                t.setGeneral(ps, genPred)
        else:
            self.triples.setGeneral(ps, genPred)

    def allTriplesGeneral(self):
        a = True
        if isinstance(self.triples, list):
            for t in self.triples:
                a = a and t.allTriplesGeneral()
        else:
            a = self.triples.allTriplesGeneral()
        return a

    def allTriplesLowSelectivity(self):
        a = True
        if isinstance(self.triples, list):
            for t in self.triples:
                a = a and t.allTriplesLowSelectivity()
        else:
            a = self.triples.allTriplesLowSelectivity()
        return a

    def instantiate(self, d):
        if isinstance(self.triples, list):
             ts = [t.instantiate(d) for t in self.triples]
             return JoinBlock(ts)
        else:
             return self.triples.instantiate(d)

    def instantiateFilter(self, d, filter_str):
        if isinstance(self.triples, list):
            ts = [t.instantiateFilter(d, filter_str) for t in self.triples]
            return JoinBlock(ts, filter_str)
        else:
            return self.triples.instantiateFilter(d, filter_str)

    def show2(self, w):
        n = nest(self.triples)
        if n:
            return aux2(n, w, " UNION ") + " ".join(map(str, self.filters))
        else:
            return " "

    def getVars(self):
        l = []
        for t in self.triples:
            l = l + t.getVars()
        return l

    def getConsts(self):
        c = []
        for t in self.triples:
            c = c + t.getConsts()
        return c

    def getPredVars(self):
        l=[]
        for t in self.triples:
            l = l + t.getPredVars()
        return l

    def includeFilter(self, f):
        for t in self.triples:
            t.includeFilter(f)

    def places(self):
        p = 0
        for e in self.triples:
            p = p + e.places()
        return p

    def constantNumber(self):
        c = 0
        for e in self.triples:
            c = c + e.constantNumber()
        return c

    def constantPercentage(self):
        return self.constantNumber()/self.places()


def nest(l):
    l0 = list(l)
    while len(l0) > 1:
        l1 = []
        while len(l0) > 1:
            x = l0.pop()
            y = l0.pop()
            l1.append((x, y))
        if len(l0) == 1:
            l1.append(l0.pop())
        l0 = l1
    if len(l0) == 1:
        return l0[0]
    else:
        return None


class JoinBlock(object):

    def __init__(self, triples, filters=[], filters_str=''):
        self.triples = triples
        self.filters = filters
        self.filters_str = filters_str

    def __repr__(self):
        r = ""
        if isinstance(self.triples, list):
            for t in self.triples:
                if isinstance(t, list):
                    r = r + " . ".join(map(str, t))
                elif t:
                    if r:
                        r = r + " . " + str(t)
                    else:
                        r = str(t)

            if len(self.filters) > 0:
                for f in self.filters:
                    r += str(f)

        else:
            r = str(self.triples)

            if len(self.filters) > 0:
                for f in self.filters:
                    r += str(f)

        return r #+ self.filters

    def setGeneral(self, ps, genPred):
        if isinstance(self.triples, list):
            for t in self.triples:
                t.setGeneral(ps, genPred)
        else:
            self.triples.setGeneral(ps, genPred)

    def allTriplesGeneral(self):
        a = True
        if isinstance(self.triples, list):
            for t in self.triples:
                a = a and t.allTriplesGeneral()
        else:
            a = self.triples.allTriplesGeneral()
        return a

    def allTriplesLowSelectivity(self):
        a = True
        if isinstance(self.triples, list):
            for t in self.triples:
                a = a and t.allTriplesLowSelectivity()
        else:
            a = self.triples.allTriplesLowSelectivity()
        return a

    def show(self, x):
        if isinstance(self.triples, list):
            joinBody = ""
            for j in self.triples:
                if isinstance(j, list):
                    if joinBody:
                        joinBody = joinBody + ". ".join(map(str, j))
                    else:
                        joinBody = joinBody + " ".join(map(str, j))
                else:
                    if joinBody:
                        joinBody = joinBody + ". " + str(j)
                    else:
                        joinBody = joinBody + " " + str(j)
            if len(self.filters) > 0:
                for f in self.filters:
                    joinBody += str(f)
            return joinBody
            #return ". ".join(map(str, self.triples)) + " ".join(map(str, self.filters)) + self.filters_str
            #n = nest(self.triples)
            #if n:
            #    return aux(n, x, " . ") + " ".join(map(str, self.filters)) + self.filters_str
            #else:
            #    return " "
        else:
            return self.triples.show(x)

    def instantiate(self, d):
        if isinstance(self.triples, list):
            ts = [t.instantiate(d) for t in self.triples]
            return JoinBlock(ts, self.filters)
        else:
            return self.triples.instantiate(d)

    def instantiateFilter(self, d, filter_str):
        if isinstance(self.triples, list):
            ts = [t.instantiateFilter(d, filter_str) for t in self.triples]
            return JoinBlock(ts, self.filters, filter_str)
        else:
            return self.triples.instantiateFilter(d, filter_str)

    def show2(self, x):
        if isinstance(self.triples, list):
            n = nest(self.triples)
            if n:
                return aux2(n, x, " . ") + str(self.filters) + self.filters_str
            else:
                return " "
        else:
            return self.triples.show2(x)

    def getVars(self):
        l = []
        if isinstance(self.triples, list):
            for t in self.triples:
                l = l + t.getVars()
        else:
            l = self.triples.getVars()
        return l

    def getConsts(self):
        c = []
        if isinstance(self.triples, list):
            for t in self.triples:
                c = c + t.getConsts()
        else:
            c = self.triples.getConsts()

        return c

    def getPredVars(self):
        l = []
        if isinstance(self.triples, list):
            for t in self.triples:
                l = l + t.getPredVars()
        else:
            l = self.triples.getPredVars()
        return l

    def includeFilter(self, f):
        for t in self.triples:
            if isinstance(t, list):
                for s in t:
                    s.include_filter(f)
            else:
                t.includeFilter(f)

    def places(self):
        p = 0
        if isinstance(self.triples, list):
            for e in self.triples:
                p = p + e.places()
        else:
            p = self.triples.places()
        return p

    def constantNumber(self):
        c = 0
        if isinstance(self.triples, list):
            for e in self.triples:
                c = c + e.constantNumber()
        else:
            c = self.triples.constantNumber()
        return c

    def constantPercentage(self):
        return self.constantNumber()/self.places()


class Filter(object):
    def __init__(self, expr):
        self.expr = expr

    def __repr__(self):
        if (self.expr.op == 'REGEX' or self.expr.op == 'sameTERM' or self.expr.op == 'langMATCHES' ):
            if (self.expr.op == 'REGEX' and self.expr.right.desc !=False):
                return "\n" + "FILTER " + self.expr.op + "(" + str(self.expr.left) + "," + self.expr.right.name + "," + self.expr.right.desc + ")"
            else:
                return "\n" + "FILTER " + self.expr.op + "(" + str(self.expr.left) + "," + str(self.expr.right) + ")"
        else:
            return "\n" + "FILTER (" + str(self.expr) + ")"

    def show(self, x):
        if self.expr.op == 'REGEX':
            if self.expr.right.desc != False:
                return "\n" + "FILTER " + self.expr.op + "(" + str(self.expr.left) + "," + self.expr.right.name + "," + self.expr.right.desc + ")"
            else:
                return "\n" + x + "FILTER regex(" + str(self.expr.left) + "," + str(self.expr.right) + ")"
        else:
            return "\n" + x + "FILTER (" + str(self.expr) + ")"

    def getVars(self):
        return self.expr.getVars()

    def getConsts(self):
        return self.expr.getConsts()

    def getVarsName(self):
        vars=[]
        for v in self.expr.getVars():
            vars.append(v[1:len(v)])
        return vars

    def getPredVars(self):
        return []

    def setGeneral(self, ps, genPred):
        return

    def places(self):
        return self.expr.places()

    def allTriplesGeneral(self):
        return False

    def allTriplesLowSelectivity(self):
        return True

    def instantiate(self, d):
        return Filter(self.expr.instantiate(d))

    def instantiateFilter(self, d, filter_str):
        return Filter(self.expr.instantiateFilter(d, filter_str))

    def constantNumber(self):
        return 1
        #return self.expr.constantNumber()

    def constantPercentage(self):
        return 0.5
        #return self.constantNumber()/self.places()


class Values(object):
    def __init__(self, vars, data_block):
        self.vars = vars
        self.data_block = data_block

    def __repr__(self):
        args_str = " ".join(map(str, self.vars))
        l_par = ''
        r_par = ''

        if len(self.vars) > 1:
            l_par = '('
            r_par = ')'
        
        data_str = ''
        for row in self.data_block:
            data_str += l_par
            for value in row:
                v = "UNDEF" if value is None else value.name
                data_str += v + " "
            data_str += r_par + '\n'

        return "\n" + "VALUES " + l_par + args_str + r_par + " { \n"  + data_str + "}"

    def show(self, x):
        args_str = " ".join(map(str, self.vars))
        
        data_str = ''
        for row in self.data_block:
            data_str += '('
            for value in row:
                v = "UNDEF" if value is None else value.name
                data_str += v + " "
            data_str += ')\n'

        return "\n" + x + "VALUES (" + args_str + ")" + "{" + data_str + "}"

    def getVars(self):
        return self.vars

    def getConsts(self):
        c = []
        for row in self.data_block:
            for arg in row:
                const = arg.getConsts() if isinstance(arg, Argument) else 'UNDEF'
                c = c + const

        return c
    
    def getVarsName(self):
        vars=[]
        for v in self.getVars():
            v = str(v)
            vars.append(v[1:len(v)])
        return vars

    def getPredVars(self):
        return []

    def setGeneral(self, ps, genPred):
        return

    def places(self):
        vars_len = len(self.vars)
        return vars_len + vars_len * len(self.data_block)

    def allTriplesGeneral(self):
        return False

    def allTriplesLowSelectivity(self):
        return True

    def instantiate(self):
        return Values(self.vars, self.data_block)

    def instantiateFilter(self, d, filter_str):
        return Values(self.vars, self.data_block)

    def constantNumber(self):
        return len(self.getConsts())

    def constantPercentage(self):
        return self.constantNumber()/self.places()


class Optional(object):
    def __init__(self, bgg):
        self.bgg = bgg

    def __repr__(self):
        return " OPTIONAL { " + str(self.bgg) + " }"

    def show(self, x):
        return x+"OPTIONAL {\n" + self.bgg.show(x + "  ") + "\n" + x + "}"

    def setGeneral(self, ps, genPred):
        self.bgg.setGeneral(ps, genPred)

    def getVars(self):
        return self.bgg.getVars()

    def getConsts(self):
        return self.bgg.getConsts()

    def getPredVars(self):
        return self.bgg.getPredVars()

    def places(self):
        return self.bgg.places()

    def allTriplesGeneral(self):
        return self.bgg.allTriplesGeneral()

    def allTriplesLowSelectivity(self):
        return self.bgg.allTriplesLowSelectivity()

    def instantiate(self, d):
        return Optional(self.bgg.instantiate(d))

    def instantiateFilter(self, d, filter_str):
        return Optional(self.bgg.instantiateFilter(d, filter_str))

    def constantNumber(self):
        return self.bgg.constantNumber()

    def constantPercentage(self):
        return self.constantNumber()/self.places()


unaryFunctor = {
     '!',
    'BOUND',
    'bound',
    'ISIRI',
    'isiri',
    'ISURI',
    'isuri',
    'ISBLANK',
    'isblank',
    'ISLITERAL',
    'isliteral',
    'STR',
    'str',
    'UCASE',
    'ucase',
    'LANG',
    'lang',
    'DATATYPE',
    'datatype',
    'xsd:double',
    'xsd:integer',
    'xsd:decimal',
    'xsd:float',
    'xsd:string',
    'xsd:boolean',
    'xsd:dateTime',
    'xsd:nonPositiveInteger',
    'xsd:negativeInteger',
    'xsd:long',
    'xsd:int',
    'xsd:short',
    'xsd:byte',
    'xsd:nonNegativeInteger',
    'xsd:unsignedInt',
    'xsd:unsignedShort',
    'xsd:unsignedByte',
    'xsd:positiveInteger',
    '<http://www.w3.org/2001/XMLSchema#integer>',
    '<http://www.w3.org/2001/XMLSchema#decimal>',
    '<http://www.w3.org/2001/XMLSchema#double>',
    '<http://www.w3.org/2001/XMLSchema#float>',
    '<http://www.w3.org/2001/XMLSchema#string>',
    '<http://www.w3.org/2001/XMLSchema#boolean>',
    '<http://www.w3.org/2001/XMLSchema#dateTime>',
    '<http://www.w3.org/2001/XMLSchema#nonPositiveInteger>',
    '<http://www.w3.org/2001/XMLSchema#negativeInteger>',
    '<http://www.w3.org/2001/XMLSchema#long>',
    '<http://www.w3.org/2001/XMLSchema#int>',
    '<http://www.w3.org/2001/XMLSchema#short>',
    '<http://www.w3.org/2001/XMLSchema#byte>',
    '<http://www.w3.org/2001/XMLSchema#nonNegativeInteger>',
    '<http://www.w3.org/2001/XMLSchema#unsignedInt>',
    '<http://www.w3.org/2001/XMLSchema#unsignedShort>',
    '<http://www.w3.org/2001/XMLSchema#unsignedByte>',
    '<http://www.w3.org/2001/XMLSchema#positiveInteger>'
    }

binaryFunctor = {
    'REGEX',
    'SAMETERM',
    'LANGMATCHES',
    'CONTAINS',
    'langMatches',
    'regex',
    'sameTerm'
    }


class Expression(object):

    def __init__(self, op, left, right):
        self.op = op
        self.left = left
        self.right = right

    def __repr__(self):
        if self.op in unaryFunctor:
            return (self.op +"("+ str(self.left) + ")")
        elif self.op in binaryFunctor:
            if (self.op == 'REGEX' and self.right.desc!=False):
                return (self.op + "("+ str(self.left) + "," + self.right.name + "," + self.right.desc + ")")
            else:
                return (self.op + "("+ str(self.left) + "," + str(self.right) + ")")
        elif self.right is None:
            return (self.op + str(self.left))
        else:
            return ("(" + str(self.left)+" "+ self.op +" "+str(self.right)+ ")")

    def getVars(self):
        if (self.op in unaryFunctor) or (self.right is None):
            return self.left.getVars()
        else:
            return self.left.getVars() + self.right.getVars()

    def getConsts(self):
        if self.op in unaryFunctor or self.right is None:
            return self.left.getConsts()
        elif self.op in binaryFunctor:
            return self.right.getConsts()
        else:
            return self.left.getConsts() + self.right.getConsts()

    def instantiate(self, d):
        return Expression(self.op, self.left.instantiate(d),
                          self.right.instantiate(d))

    def instantiateFilter(self, d, filter_str):
        return Expression(self.op, self.left.instantiateFilter(d, filter_str),
                          self.right.instantiateFilter(d, filter_str))

    def allTriplesGeneral(self):
        return False

    def allTriplesLowSelectivity(self):
        return True

    def setGeneral(self, ps, genPred):
        return

    def places(self):
        if ((self.op in unaryFunctor)or (self.op == 'REGEX' and self.right.desc ==False)):
            return self.left.places()
        else:
            return self.left.places() + self.right.places()

    def constantNumber(self):
        if ((self.op in unaryFunctor) or (self.op == 'REGEX' and self.expr.desc ==False)):
            return self.left.constantNumber()
        else:
            return self.left.constantNumber() + self.right.constantNumber()

    def constantPercentage(self):
        return self.constantNumber()/self.places()


class Triple(object):
    def __init__(self, subject, predicate, theobject):
        self.subject = subject
        self.predicate = predicate
        self.theobject = theobject
        self.isGeneral = False

    def __repr__(self):
        return ("\n        " + self.subject.name + " " + self.predicate.name + " "    + str(self.theobject))

    def setGeneral(self, ps, genPred):
        self.isGeneral = (getUri(self.predicate, ps) in genPred)

    def __eq__(self, other):

        return ((self.subject == other.subject) and
                (self.predicate == other.predicate) and
                (self.theobject == other.theobject))

    def __hash__(self):
        return hash((self.subject,self.predicate,self.theobject))

    def allTriplesGeneral(self):
        return self.isGeneral

    #Modified 17-12-2013. General predicates are not considered to decide if the triple is selective or not
    def allTriplesLowSelectivity(self):
        return ((not self.predicate.constant)
                #or ((self.isGeneral) and (not self.subject.constant)
                 or ((not self.subject.constant)
                    and (not self.theobject.constant))
                 # or
                 #   # added 09/03/2018
                 #   (self.theobject.constant and
                 #   self.predicate.constant and
                 #   ('rdf:type' in self.predicate.name or 'a' == self.predicate.name or
                 #     'http://www.w3.org/1999/02/22-rdf-syntax-ns#type' in self.predicate.name))
                )

    def show(self, x):
        return x+self.subject.name+" " + self.predicate.name + " " + str(self.theobject)

    def getVars(self):
        l = []
        if not self.subject.constant:
            l.append(self.subject.name)
        if not self.predicate.constant:
            l.append(self.predicate.name)
        if not self.theobject.constant:
            l.append(self.theobject.name)
        return l

    def getConsts(self):
        c = []
        if self.subject.constant:
            c.append(self.subject.name)
        if self.theobject.constant:
            c.append(self.theobject.name)
        return c

    def getPredVars(self):
        l = []
        if not self.predicate.constant:
            l.append(self.predicate.name)
        return l

    def places(self):
        return 3

    def instantiate(self, d):
        sn = self.subject.name.lstrip('?$')
        pn = self.predicate.name.lstrip('?$')
        on = self.theobject.name.lstrip('?$')
        if (not self.subject.constant) and (sn in d):
            s = Argument(d[sn], True)
        else:
            s = self.subject
        if (not self.predicate.constant) and (pn in d):
            p = Argument(d[pn], True)
        else:
            p = self.predicate
        if (not self.theobject.constant) and (on in d):
            o = Argument(d[on], True)
        else:
            o = self.theobject
        return Triple(s, p, o)

    def instantiateFilter(self, d, filter_str):
        return Triple(self.subject, self.predicate, self.theobject)

    def constantNumber(self):
        n = 0
        if self.subject.constant:
            n = n + 1
        if self.predicate.constant:
            n = n + 1
        if self.theobject.constant:
            n = n + 1
        return n

    def constantPercentage(self):
        return self.constantNumber()/self.places()


class Argument(object):

    def __init__(self, name, constant, desc=False, datatype=None, lang=None):
        self.name = name if not constant else urllib.parse.quote(name, safe='<:/#>"\'^')
        self.constant = constant
        self.desc = desc
        self.datatype = datatype
        self.lang = lang

    def __repr__(self):
        arg = self.name
        if self.datatype is not None:
            arg += "^^" + self.datatype
        if self.lang is not None:
            arg += "@" + self.lang

        return arg

    def __str__(self):
        arg = self.name
        if self.datatype is not None:
            arg += "^^" + self.datatype
        elif self.lang is not None:
            arg += "@" + self.lang

        return arg

    def __eq__(self, other):
        if isinstance(other, str):
            other = Argument(other, None)
        return self.name == other.name and self.constant == other.constant

    def __ne__(self, other):
        return not self == other

    def __hash__(self):
        return hash((self.name, self.constant))

    def getVars(self):
        if self.constant:
            return []
        else:
            return [self.name]

    def getConsts(self):

        if self.constant:
            n = self.name
            if self.datatype is not None:
                n = n + "^^" + self.datatype
            elif self.lang is not None:
                n = n + "@" + self.lang
            return [n]
        else:
            return []

    def places(self):
        return 1

    def constantNumber(self):
        n = 0
        if self.constant:
            n = n + 1
        return n

    def constantPercentage(self):
        return self.constantNumber()/self.places()


def readGeneralPredicates(fileName):
    f = open(fileName, 'r')
    l = []
    l0 = f.readline()
    while not l0 == '':
        l0 = l0.rstrip('\n')
        l.append(l0)
        l0 = f.readline()
    f.close()
    return l


def getUri(p, prefs):
    if p.datatype or p.lang:
        return p.name + ("^^" + p.datatype if p.datatype else "") + ("@" + p.lang if p.lang else "")
    if "'" in p.name or '"' in p.name:
        return p.name
    hasPrefix = prefix(p)
    if hasPrefix:
        (pr, su) = hasPrefix
        n = prefs[pr]
        n = n[:-1]+su+">"
        return n
    return p.name


def prefix(p):
    s = p.name
    pos = s.find(":")
    if (not (s[0] == "<")) and pos > -1:
        return (s[0:pos].strip(), s[(pos+1):].strip())

    return None


def getPrefs(ps):
    prefDict = dict()
    for p in ps:
        pos = p.find(":")
        c = p[0:pos].strip()
        v = p[(pos+1):len(p)].strip()
        prefDict[c] = v
    return prefDict



"""
Created on Jul 10, 2011

Implements the structures used by the Blocking Operators.

@author: Maribel Acosta Deibe
"""


class Record(object):
    """
    Represents a structure that is inserted into the hash table.
    It is composed by a tuple, ats (arrival timestamp) and
    dts (departure timestamp).
    """

    def __init__(self, tuple, ats, dts):
        self.tuple = tuple
        self.ats = ats
        self.dts = dts


class Partition(object):
    """
    Represents a bucket of the hash table.
    It is composed by a list of records, and a list of timestamps
    of the form {DTSlast, ProbeTS}
    """

    def __init__(self):
        self.records = []  # List of records
        self.timestamps = []  # List of the form {DTSlas, ProbeTS}


class Table(object):
    """
    Represents a hash table.
    It is composed by a list of partitions (buckets) of size n,
    where n is specified in "size".
    """

    def __init__(self):
        self.size = 3
        self.partitions = [Partition() for x in range(self.size)]

    def getSize(self):
        return self.size

    def insertRecord(self, i, value):
        self.partitions[i].records.append(value)



from __future__ import annotations

__author__ = 'Kemele M. Endris and Philipp D. Rohde'

import abc
import json
import os
import time
from base64 import b64encode

import requests


class Config(object):
    def __init__(self, configfile=None, json_data=None):
        self.configfile = configfile
        self.metadata = {}
        self.predidx = {}
        self.predwrapidx = {}
        self.endpoints = {}
        if configfile is not None or json_data is not None:
            if json_data is not None:
                self.metadata = json_data
            self.metadata = self.getAll()
            if self.metadata is None:
                self.metadata = {}
            self.predidx = self.createPredicateIndex()
            self.predwrapidx = self.createPredicateWrapperIndex()
            self.endpoints = self.getEndpoints()

    @abc.abstractmethod
    def getAll(self):
        return

    def getEndpoints(self):
        endpoints = {}
        for m in self.metadata:
            wrappers = self.metadata[m]['wrappers']
            for w in wrappers:
                if w['url'] not in endpoints:
                    endpoints[w['url']] = w['urlparam']
        return endpoints

    def setEndpointToken(self, endpoint, token, valid_until):
        self.endpoints[endpoint]['token'] = token
        self.endpoints[endpoint]['valid_until'] = valid_until

    @staticmethod
    def __get_auth_token(server, username, password):
        payload = 'grant_type=client_credentials&client_id=' + username + '&client_secret=' + password
        headers = {'Content-Type': 'application/x-www-form-urlencoded'}

        start = time.time()
        response = requests.request('POST', server, headers=headers, data=payload)
        if response.status_code != 200:
            raise Exception(str(response.status_code) + ': ' + response.text)
        return response.json()['access_token'], start + response.json()['expires_in']

    def get_auth(self, endpoint):
        params = self.endpoints.get(endpoint, None)
        if params is not None and 'username' in params and 'password' in params:
            if 'keycloak' in params:
                valid_token = False
                if 'token' in params and 'valid_until' in params:
                    current = time.time()
                    if params['valid_until'] > current:
                        valid_token = True

                if valid_token:
                    token = params['token']
                else:
                    token, valid_until = self.__get_auth_token(params['keycloak'], params['username'], params['password'])
                    self.setEndpointToken(endpoint, token, valid_until)
                return 'Bearer ' + token
            else:
                credentials = params['username'] + ':' + params['password']
                return 'Basic ' + b64encode(credentials.encode()).decode()
        return None

    def createPredicateIndex(self):
        pidx = {}
        for m in self.metadata:
            preds = self.metadata[m]['predicates']
            for p in preds:
                if p['predicate'] not in pidx:
                    pidx[p['predicate']] = set()
                    pidx[p['predicate']].add(m)
                else:
                    pidx[p['predicate']].add(m)

        return pidx

    def createPredicateWrapperIndex(self):
        idx = {}
        for m in self.metadata:
            wrappers = self.metadata[m]['wrappers']
            for wrapper in wrappers:
                preds = wrapper['predicates']
                for pred in preds:
                    if pred not in idx:
                        idx[pred] = set()
                        idx[pred].add(wrapper['url'])
                    else:
                        idx[pred].add(wrapper['url'])
        return idx

    def findbypreds(self, preds):
        res = []
        for p in preds:
            if p in self.predidx:
                res.append(self.predidx[p])
        if len(res) != len(preds):
            return []
        for r in res[1:]:
            res[0] = res[0].intersection(r)

        mols = list(res[0])
        return mols

    def find_preds_per_mt(self, preds):
        res = {}
        for p in preds:
            if p in self.predidx:
                for m in self.predidx[p]:
                    res.setdefault(m, []).append(p)

        respreds = []
        for m in res:
            respreds.extend(res[m])
        if len(set(list(respreds))) != len(preds):
            return {}

        return res

    def findbypred(self, pred):
        mols = []
        for m in self.metadata:
            mps = [pm['predicate'] for pm in self.metadata[m]['predicates']]
            if pred in mps:
                mols.append(m)
        return mols

    def findMolecule(self, molecule):
        if molecule in self.metadata:
            return self.metadata[molecule]
        else:
            return None

    def saveToFile(self, path):
        with open(path, 'w', encoding='utf8') as output_file:
            json.dump([self.metadata[m] for m in self.metadata], output_file, indent=2)


class ConfigFile(Config):
    def __init__(self, configfile):
        if os.path.isfile(configfile):
            super().__init__(configfile=configfile)
        else:
            super().__init__()

    def getAll(self):
        return self.read_json_file(self.configfile)

    @staticmethod
    def read_json_file(configfile):
        try:
            with open(configfile, 'r', encoding='utf8') as f:
                mts = json.load(f)

                meta = {}
                for m in mts:
                    if m['rootType'] in meta:
                        # linkedTo
                        links = meta[m['rootType']]['linkedTo']
                        links.extend(m['linkedTo'])
                        meta[m['rootType']]['linkedTo'] = list(set(links))

                        # predicates
                        preds = meta[m['rootType']]['predicates']
                        mpreds = m['predicates']
                        ps = {p['predicate']: p for p in preds}
                        for p in mpreds:
                            if p['predicate'] in ps and len(p['range']) > 0:
                                ps[p['predicate']]['range'].extend(p['range'])
                                ps[p['predicate']]['range'] = list(set(ps[p['predicate']]['range']))
                            else:
                                ps[p['predicate']] = p

                        meta[m['rootType']]['predicates'] = []
                        for p in ps:
                            meta[m['rootType']]['predicates'].append(ps[p])

                        # wrappers
                        wraps = meta[m['rootType']]['wrappers']
                        wrs = {w['url']+w['wrapperType']: w for w in wraps}
                        mwraps = m['wrappers']
                        for w in mwraps:
                            key = w['url'] + w['wrapperType']
                            if key in wrs:
                                wrs[key]['predicates'].extend(wrs['predicates'])
                                wrs[key]['predicates'] = list(set(wrs[key]['predicates']))

                        meta[m['rootType']]['wrappers'] = []
                        for w in wrs:
                            meta[m['rootType']]['wrappers'].append(wrs[w])
                    else:
                        meta[m['rootType']] = m
            f.close()
            return meta
        except Exception as e:
            print("Exception while reading molecule templates file:", e)
            return None


class MTCreationConfig(Config):
    def __init__(self):
        super().__init__()
        self.endpoints = {}

    def addEndpoint(self, url, params: dict = None):
        if url not in self.endpoints:
            self.endpoints[url] = params

    def setEndpoints(self, endpoints: list | dict):
        self.endpoints = {}
        if isinstance(endpoints, list):
            [self.addEndpoint(e) for e in endpoints]
        elif isinstance(endpoints, dict):
            [self.addEndpoint(key, value) for key, value in endpoints.items()]

    def getAll(self):
        return None


class JSONConfig(Config):
    def __init__(self, json_data):
        super().__init__(json_data=json_data)

    def getAll(self):
        meta = {}
        for m in self.metadata:
            meta[m['rootType']] = m
        return meta



"""
NestedHashJoin.py

Implements a depending operator, similar to block nested join and symmetric
hash join

Autor: Maribel Acosta
Autor: Maria-Esther Vidal
Autor: Gabriela Montoya
Date: January 29th, 2014
"""

from multiprocessing import Queue, Process
from time import time
import string, sys
from queue import Empty
from DeTrusty.Operators.Join import Join
from DeTrusty.Sparql.Parser import queryParser as qp
from .OperatorStructures import Table, Partition, Record
from .NestedHashJoin import NestedHashJoin

WINDOW_SIZE = 20


class NestedHashJoinFilter(Join):

    def __init__(self, vars):
        self.left_table = dict()
        self.right_table = dict()
        self.qresults    = Queue()
        self.vars        = vars
        self.left_queue = Queue()
        self.right_operator = None
        self.qresults = Queue()

    def instantiate(self, d):
        newvars = self.vars - set(d.keys())
        return NestedHashJoin(newvars)

    def instantiateFilter(self, d, filter_str):
        newvars = self.vars - set(d)
        return NestedHashJoin(newvars)

    def execute(self, left_queue, right_operator, out, processqueue=Queue()):
        self.left_queue = left_queue
        self.right_operator = right_operator
        self.qresults = out
        # print "right_operator", right_operator
        tuple1 = None
        tuple2 = None
        right_queues = dict()
        filter_bag = []
        count = 0
        while (not (tuple1 == "EOF") or (len(right_queues) > 0)):
            try:
                tuple1 = self.left_queue.get(False)
                # Try to get and process tuple from left queue
                if not (tuple1 == "EOF"):
                    # tuple1 = self.left_queue.get(False)
                    # print "tuple1: "+str(tuple1)
                    instance = self.probeAndInsert1(tuple1, self.right_table,
                                                    self.left_table, time())
                    # print "sali de probe and insert 1 con tuple", tuple1
                    if instance:  # the join variables have not been used to
                        # instanciate the right_operator
                        filter_bag.append(tuple1)
                    # print "filter_bag", len(filter_bag)

                    if len(filter_bag) >= WINDOW_SIZE:
                        new_right_operator = self.makeInstantiation(filter_bag,
                                                                    self.right_operator)
                        # print "Here in makeInstantation with filter"
                        # resource = self.getResource(tuple1)
                        queue = Queue()
                        right_queues[count] = queue
                        new_right_operator.execute(queue)
                        filter_bag = []
                        count = count + 1

                else:
                    if (len(filter_bag) > 0):
                        # print "here", len(filter_bag), filter_bag
                        new_right_operator = self.makeInstantiation(filter_bag,
                                                                    self.right_operator)
                        # resource = self.getResource(tuple1)
                        queue = Queue()
                        right_queues[count] = queue
                        new_right_operator.execute(queue)
                        filter_bag = []
                        count = count + 1

            except Empty:
                pass
            except Exception as e:
                # print "Unexpected error:", sys.exc_info()[0]
                # print e
                pass

            toRemove = []  # stores the queues that have already received all its tuples
            # print "right_queues", right_queues
            for r in right_queues:
                try:
                    q = right_queues[r]
                    tuple2 = None
                    while (tuple2 != "EOF"):
                        tuple2 = q.get(False)

                        if (tuple2 == "EOF"):
                            toRemove.append(r)
                        else:
                            resource = self.getResource(tuple2)
                            for v in self.vars:
                                del tuple2[v]
                            # print "new tuple2", tuple2
                            self.probeAndInsert2(resource, tuple2, self.left_table, self.right_table, time())
                except Exception:
                    # This catch:
                    # Empty: in tuple2 = self.right.get(False), when the queue is empty.
                    # TypeError: in att = att + tuple[var], when the tuple is "EOF".
                    # print "Unexpected error:", sys.exc_info()
                    pass

            for r in toRemove:
                if r in right_queues:
                    del right_queues[r]
        # Put EOF in queue and exit.
        self.qresults.put("EOF")
        return

    def getResource(self, tuple):
        resource = ''
        for var in self.vars:
            val = tuple[var]
            if "^^<" in val:
                val = val[:val.find('^^<')]
            resource = resource + val
        return resource

    def makeInstantiation(self, filter_bag, operators):
        filter_str = ''
        new_vars = ['?' + v for v in self.vars]  # TODO: this might be $
        filter_str = " . ".join(map(str, operators.tree.service.filters))
        # print "making instantiation join filter", filter_bag
        # When join variables are more than one: FILTER ( )
        if len(self.vars) >= 1:
            filter_str += ' . FILTER (__expr__)'

            or_expr = []
            for tuple in filter_bag:
                and_expr = []
                for var in self.vars:
                    # aux = "?" + var + "==" + tuple[var]
                    v = tuple[var]
                    if v.find("http") == 0:  # uris must be passed between < .. >
                        v = "<" + v + ">"
                        v = "?" + var + "=" + v
                        and_expr.append(v)
                    else:
                        if '^^<' not in v:
                            v = '"' + v + '"'
                            vf = "str(?" + var + ")=" + v  # + " || " + "?" + var + "=" + v + "^^<http://www.w3.org/2001/XMLSchema#string>)"
                            and_expr.append(vf)
                        else:
                            loc = v.find('^^<')
                            vf = '"' + v[:loc] + '"' + v[loc:]
                            v = "(?" + var + "=" + vf + ' || ' + "?" + var + '="' + v[:loc] + '")'
                            and_expr.append(v)
                if len(and_expr) > 1:
                    or_expr.append('(' + ' && '.join(and_expr) + ')')
                else:
                    or_expr.append(' && '.join(and_expr))
            filter_str = filter_str.replace('__expr__', ' || '.join(or_expr))
        new_operator = operators.instantiateFilter(set(new_vars), filter_str)
        # print "type(new_operator)", type(new_operator)

        return new_operator

    def makeInstantiationX(self, filter_bag, operators):
        filter_str = ''

        import copy
        operator = copy.copy(operators)
        # print "||||||||||||||||||||||||||||||"
        # print 'operator type ', type(operators)
        # print operator
        # operator = operator.copy()
        # print "TreePlan" in operator.__class__.__name__

        if "TreePlan" in operator.__class__.__name__:
            if operator.left and operator.left.__class__.__name__ == "TreePlan":
                operator.left = copy.copy(operators.left)
                if operator.left.left and operator.left.left.__class__.__name__ == "IndependentOperator":
                    operator.left.left = copy.copy(operators.left.left)
                    filter_str = " . ".join(map(str, operator.left.left.tree.service.filters))
                    new_vars = ['?' + v for v in self.vars]
                    if len(self.vars) >= 1:
                        filter_str += ' . FILTER (__expr__)'

                        or_expr = []
                        for tuple in filter_bag:
                            and_expr = []
                            for var in self.vars:
                                # aux = "?" + var + "==" + tuple[var]
                                v = str(tuple[var])
                                if v.find("http") == 0:  # uris must be passed between < .. >
                                    v = "<" + v + ">"
                                else:
                                    v = '"' + v + '"'
                                v = "?" + var + "=" + v
                                and_expr.append(v)

                            or_expr.append('(' + ' && '.join(and_expr) + ')')
                        filter_str = filter_str.replace('__expr__', ' || '.join(or_expr))
                    operator.left.left = operator.left.left.instantiateFilter(set(new_vars), filter_str)
                if operator.left.right and operator.left.right.__class__.__name__ == "IndependentOperator":
                    operator.left.right = copy.copy(operators.left.right)
                    filter_str = " . ".join(map(str, operator.left.right.tree.service.filters))
                    new_vars = ['?' + v for v in self.vars]
                    if len(self.vars) >= 1:
                        filter_str += ' . FILTER (__expr__)'

                        or_expr = []
                        for tuple in filter_bag:
                            and_expr = []
                            for var in self.vars:
                                # aux = "?" + var + "==" + tuple[var]
                                v = str(tuple[var])
                                if v.find("http") == 0:  # uris must be passed between < .. >
                                    v = "<" + v + ">"
                                else:
                                    v = '"' + v + '"'
                                v = "?" + var + "=" + v
                                and_expr.append(v)

                            or_expr.append('(' + ' && '.join(and_expr) + ')')
                        filter_str = filter_str.replace('__expr__', ' || '.join(or_expr))
                    operator.left.right = operator.left.right.instantiateFilter(set(new_vars), filter_str)
            if operator.left and operator.left.__class__.__name__ == "IndependentOperator":
                operator.left = copy.copy(operators.left)
                filter_str = " . ".join(map(str, operator.left.tree.service.filters))
                new_vars = ['?' + v for v in self.vars]
                if len(self.vars) >= 1:
                    filter_str += ' . FILTER (__expr__)'

                    or_expr = []
                    for tuple in filter_bag:
                        and_expr = []
                        for var in self.vars:
                            # aux = "?" + var + "==" + tuple[var]
                            v = str(tuple[var])
                            if v.find("http") == 0:  # uris must be passed between < .. >
                                v = "<" + v + ">"
                            else:
                                v = '"' + v + '"'
                            v = "?" + var + "=" + v
                            and_expr.append(v)

                        or_expr.append('(' + ' && '.join(and_expr) + ')')
                    filter_str = filter_str.replace('__expr__', ' || '.join(or_expr))
                operator.left = operator.left.instantiateFilter(set(new_vars), filter_str)
                # print "op left", operator.left
            if operator.right and operator.right.__class__.__name__ == "TreePlan":
                operator.right = copy.copy(operators.right)
                if operator.right.left.__class__.__name__ == "IndependentOperator":
                    operator.right.left = copy.copy(operators.right.left)
                    filter_str = " . ".join(map(str, operator.right.left.tree.service.filters))
                    new_vars = ['?' + v for v in self.vars]
                    if len(self.vars) >= 1:
                        filter_str += ' . FILTER (__expr__)'

                        or_expr = []
                        for tuple in filter_bag:
                            and_expr = []
                            for var in self.vars:
                                # aux = "?" + var + "==" + tuple[var]
                                v = str(tuple[var])
                                if v.find("http") == 0:  # uris must be passed between < .. >
                                    v = "<" + v + ">"
                                else:
                                    v = '"' + v + '"'
                                v = "?" + var + "=" + v
                                and_expr.append(v)

                            or_expr.append('(' + ' && '.join(and_expr) + ')')
                        filter_str = filter_str.replace('__expr__', ' || '.join(or_expr))
                    operator.right.left = operator.right.left.instantiateFilter(set(new_vars), filter_str)
                if operator.right.right.__class__.__name__ == "IndependentOperator":
                    operator.right.right = copy.copy(operators.right.right)
                    filter_str = " . ".join(map(str, operator.right.right.tree.service.filters))
                    new_vars = ['?' + v for v in self.vars]
                    if len(self.vars) >= 1:
                        filter_str += ' . FILTER (__expr__)'

                        or_expr = []
                        for tuple in filter_bag:
                            and_expr = []
                            for var in self.vars:
                                # aux = "?" + var + "==" + tuple[var]
                                v = str(tuple[var])
                                if v.find("http") == 0:  # uris must be passed between < .. >
                                    v = "<" + v + ">"
                                else:
                                    v = '"' + v + '"'
                                v = "?" + var + "=" + v
                                and_expr.append(v)

                            or_expr.append('(' + ' && '.join(and_expr) + ')')
                        filter_str = filter_str.replace('__expr__', ' || '.join(or_expr))
                    operator.right.right = operator.right.right.instantiateFilter(set(new_vars), filter_str)
            # print type(operator)
            if operator.right and operator.right.__class__.__name__ == "IndependentOperator":
                operator.right = copy.copy(operators.right)
                filter_str = " . ".join(map(str, operator.right.tree.service.filters))
                new_vars = ['?' + v for v in self.vars]
                if len(self.vars) >= 1:
                    filter_str += ' . FILTER (__expr__)'

                    or_expr = []
                    for tuple in filter_bag:
                        and_expr = []
                        for var in self.vars:
                            # aux = "?" + var + "==" + tuple[var]
                            v = str(tuple[var])
                            if v.find("http") == 0:  # uris must be passed between < .. >
                                v = "<" + v + ">"
                            else:
                                v = '"' + v + '"'
                            v = "?" + var + "=" + v
                            and_expr.append(v)

                        or_expr.append('(' + ' && '.join(and_expr) + ')')
                    filter_str = filter_str.replace('__expr__', ' || '.join(or_expr))
                operator.right = operator.right.instantiateFilter(set(new_vars), filter_str)

            return operator
        else:
            # for t in operator.tree.service.filters:
            filter_str = " . ".join(map(str, operator.tree.service.filters))
            # print "||||||||||||||||||||||||||||||"
            new_vars = ['?' + v for v in self.vars]  # TODO: this might be $
            # print "NestedHashJoinFilter: makeInstantiattion ",new_vars
            # print "making instantiation join filter", filter_bag
            # When join variables are more than one: FILTER ( )
            if len(self.vars) >= 1:
                filter_str += ' . FILTER (__expr__)'

                or_expr = []
                for tuple in filter_bag:
                    and_expr = []
                    for var in self.vars:
                        # aux = "?" + var + "==" + tuple[var]
                        v = str(tuple[var])
                        if v.find("http") == 0:  # uris must be passed between < .. >
                            v = "<" + v + ">"
                        else:
                            v = '"' + v + '"'
                        v = "?" + var + "=" + v
                        and_expr.append(v)

                    or_expr.append('(' + ' && '.join(and_expr) + ')')
                filter_str = filter_str.replace('__expr__', ' || '.join(or_expr))

            # When join variables is a singleton: FILTER ?var IN ()
            # else:
            #
            #    filter_str = ' . FILTER ?' + list(self.vars)[0] + ' IN (__expr__)'
            #
            #    expr = []
            #    for tuple in filter_bag:
            #        v = tuple[list(self.vars)[0]]
            #        if string.find(v, "http") == 0: # uris must be passed between < .. >
            #            v = "<"+v+">"
            #        expr.append(v)
            #    filter_str = filter_str.replace('__expr__', ' , '.join(expr))

            # print "NestedHashJoinFilter: makeInstantiattion filter expression:", filter_str
            # print "NestedHashJoinFilter: makeInstantiattion type(operator)", type(operator)
            # print "////////////////NEW OP////////////////////////////////////"

            new_operator = operator.instantiateFilter(set(new_vars), filter_str)
            # print "NestedHashJoinFilter: makeInstantiattion type(new_operator)", type(new_operator)

            # print new_operator
            # print "////////////////////////////////////////////////////"
            return new_operator

    def probeAndInsert1(self, tuple, table1, table2, time):
        #print "in probeAndInsert1", tuple
        record = Record(tuple, time, 0)
        r = self.getResource(tuple)
        #print "resource", r, tuple
        if r in table1:
            records =  table1[r]
            for t in records:
                if t.ats > record.ats:
                    continue
                x = t.tuple.copy()
                x.update(tuple)
                self.qresults.put(x)
        p = table2.get(r, [])
        i = (p == [])
        p.append(record)
        table2[r] = p
        return i

    def probeAndInsert2(self, resource, tuple, table1, table2, time):
        #print "probeAndInsert2", resource, tuple
        record = Record(tuple, time, 0)
        if resource in table1:
            records =  table1[resource]
            for t in records:
                if t.ats > record.ats:
                    continue
                x = t.tuple.copy()
                x.update(tuple)
                self.qresults.put(x)
        p = table2.get(resource, [])
        p.append(record)
        table2[resource] = p



import json
import urllib.parse
import urllib.request

from DeTrusty.Logger import get_logger

logger = get_logger('DeTrusty.Wrapper.RDFWrapper')


def contact_source(server, query, queue, config, limit=-1):
    # Contacts the datasource (i.e. real endpoint).
    # Every tuple in the answer is represented as Python dictionaries
    # and is stored in a queue.
    logger.info("Contacting endpoint: " + server)
    b = None
    cardinality = 0

    if limit == -1:
        b, cardinality = contact_source_aux(server, query, queue, config)
    else:
        # Contacts the datasource (i.e. real endpoint) incrementally,
        # retrieving partial result sets combining the SPARQL sequence
        # modifiers LIMIT and OFFSET.

        # Set up the offset.
        offset = 0

        while True:
            query_copy = query + " LIMIT " + str(limit) + " OFFSET " + str(offset)
            b, card = contact_source_aux(server, query_copy, queue, config)
            cardinality += card
            if card < limit:
                break

            offset = offset + limit

    # Close the queue
    queue.put("EOF")
    return b, cardinality


def contact_source_aux(server, query, queue, config=None):
    # Setting variables to return.
    b = None
    cardinality = 0

    payload = {'query': query, 'format': 'JSON'}
    headers = {"User-Agent":
                   "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/70.0.3538.77 Safari/537.36",
               "Accept": "application/sparql-results+json"}

    auth = config.get_auth(server)
    if auth is not None:
        headers['Authorization'] = auth

    try:
        data = urllib.parse.urlencode(payload)
        data = data.encode('utf-8')
        req = urllib.request.Request(server, data, headers)
        with urllib.request.urlopen(req) as response:
            res = json.loads(response.read().decode('utf8'))  # TODO: does this need another try block?
            if isinstance(res, dict):
                b = res.get('boolean', None)

                if 'results' in res:
                    # print "raw results from endpoint", res
                    for x in res['results']['bindings']:
                        for key, props in x.items():
                            # Handle typed-literals and language tags
                            suffix = ''
#                            if props['type'] == 'typed-literal':
#                                if isinstance(props['datatype'], bytes):
#                                    suffix = "^^<" + props['datatype'].decode('utf-8') + ">"
#                                else:
#                                    suffix = "^^<" + props['datatype'] + ">"
#                            elif "xml:lang" in props:
#                                suffix = '@' + props['xml:lang']
                            try:
                                if isinstance(props['value'], bytes):
                                    x[key] = props['value'].decode('utf-8') + suffix
                                else:
                                    x[key] = props['value'] + suffix
                            except:
                                x[key] = props['value'] + suffix

                        queue.put(x)
                        cardinality += 1
                    # Every tuple is added to the queue.
                    # for elem in reslist:
                    # print elem
                    # queue.put(elem)

            else:
                logger.error("the source " + str(server) + " answered in " + res.getheader(
                    "content-type") + " format, instead of"
                      + " the JSON format required, then that answer will be ignored")
    except Exception as e:
        logger.error("Exception while sending request to " + str(server) + " - msg: " + str(e) + " - query: " + str(query))
        return None, -2  # indicating an error during the query execution

    return b, cardinality



import abc
from multiprocessing import Queue


class Optional(object):
    __metaclass__ = abc.ABCMeta

    @abc.abstractmethod
    def execute(self, left, right, out, processqueue=Queue()):
        return

    @abc.abstractmethod
    def instantiate(self, d):
        return

    def getCardinality(self, l, r):
        c1 = l.getCardinality()
        c2 = r.getCardinality()
        jc1 = l.getJoinCardinality(self.vars)
        jc2 = r.getJoinCardinality(self.vars)
        return c1 * c2 * (1 / max(jc1, jc2))

    def getJoinCardinality(self, l, r, vars):
        return max(l.getJoinCardinality(vars), r.getJoinCardinality(vars))



"""
Created on Jul 10, 2011

Implements a XJoin operator.
The intermediate results are represented as queues.

@author: Maribel Acosta Deibe
"""

from multiprocessing import Queue
from queue import Empty
from tempfile import NamedTemporaryFile
from threading import Timer
from random import randint
from os import remove
from DeTrusty.Operators.Join import Join
from .OperatorStructures import Table, Partition, Record, FileDescriptor, isOverlapped


class XJoin(Join):

    def __init__(self, vars):
        self.left_table  = Table()
        self.right_table = Table()
        self.qresults    = Queue()
        self.vars        = vars
        self.timestamp   = 0

        # Second stage settings
        self.timeoutSecondStage  = 4
        self.leftSourceBlocked   = False
        self.rightSourceBlocked  = False

        # Main memory settings
        self.memorySize   = 10        # Represents the main memory size (# tuples).OLD:Represents the main memory size (in KB).
        self.memory_left  = 0
        self.memory_right = 0
        self.fileDescriptor_left = {}
        self.fileDescriptor_right = {}

    def instantiate(self, d):
        newvars = self.vars - set(d.keys())
        return XJoin(newvars)

    def execute(self, left, right, out, processqueue=Queue()):
        # Executes the XJoin.
        self.left     = left
        self.right    = right
        self.qresults = out

        # Initialize tuples.
        tuple1 = None
        tuple2 = None

        # Create alarms (timers) to stage 2.
        t1 = Timer(float("inf"), self.leftBlocked)
        t2 = Timer(float("inf"), self.rightBlocked)

        # Get the tuples from the queues.
        while (not(tuple1 == "EOF") or not(tuple2 == "EOF")):

            if (self.memory_left + self.memory_right >= self.memorySize):
                self.flushPartition()

            # Try to get and process tuple from left queue.
            if (not(tuple1 == "EOF")):
                try:
                    tuple1 = self.left.get(False)
                    self.leftSourceBlocked = False
                    self.timestamp += 1
                    t1.cancel()
                    t1 = Timer(self.timeoutSecondStage, self.leftBlocked)
                    t1.start()
                    self.stage1(tuple1, self.left, self.left_table, self.right_table)
                    self.memory_left += 1
                except Empty:
                    # Empty: in tuple1 = self.left.get(False), when the queue is empty.
                    #self.stage2()
                    pass
                except TypeError:
                    # TypeError: in att = att + tuple[var], when the tuple is "EOF".
                    pass

            # Try to get and process tuple from right queue.
            if (not(tuple2 == "EOF")):
                try:
                    tuple2 = self.right.get(False)
                    self.rightSourceBlocked = False
                    self.timestamp += 1
                    t2.cancel()
                    t2 = Timer(self.timeoutSecondStage, self.rightBlocked)
                    t2.start()
                    self.stage1(tuple2, self.right, self.right_table, self.left_table)
                    self.memory_right += 1
                except Empty:
                    # Empty: in tuple1 = self.right.get(False), when the queue is empty.
                    #self.stage2()
                    pass
                except TypeError:
                    # TypeError: in att = att + tuple[var], when the tuple is "EOF".
                    pass

        # Turn off alarms to stage 2.
        t1.cancel()
        t2.cancel()
        # Perform the last probes.
        self.stage3()

    def stage1(self, tuple, input, table1, table2):
        # Stage 1: While both sources are sending data.

        # Get the attribute(s) to apply hash.
        att = ''
        for var in self.vars:
            att = att + tuple[var]
        i = hash(att) % table1.size;

        # Insert record in partition.
        #record = Record(tuple, time(), 0)
        record = Record(tuple, self.timestamp, float("inf"))
        table1.insertRecord(i, record)

        # Probe the record against its partition in the other table.
        self.probe(record, table2.partitions[i], self.vars)

    def leftBlocked(self):
        self.leftSourceBlocked = True
        self.stage2()

    def rightBlocked(self):
        self.rightSourceBlocked = True
        self.stage2()

    def stage2(self):
        # Stage 2: When both sources stops sending data.
        if (self.leftSourceBlocked and self.rightSourceBlocked):
            # Timestamp of the secondStage
            probeTS = self.timestamp

        while (self.leftSourceBlocked and self.rightSourceBlocked):
            # Choose a random victim from main memory.
            i = randint(0, (self.left_table.size + self.right_table.size)-1)

            if (i < self.left_table.size):
                partition = self.left_table.partitions[i]
                file_descriptor = self.fileDescriptor_right
            else:
                i = i % self.right_table.size
                partition = self.right_table.partitions[i]
                file_descriptor = self.fileDescriptor_left

            # Check if the partition has been flushed and has records in main memory.
            if ((file_descriptor.has_key(i)) and (len(partition.records)>0)):
                dtsLast = 0

                for record in partition.records:
                    if (record.ats <= probeTS):
                        dtsLast = self.probeMainSecondaryMem(record, file_descriptor, i, probeTS)

                # Update the file descriptor.
                file_descriptor[i].timestamps.add((dtsLast, probeTS))

    def stage3(self):
        # Stage 3: When both sources sent all the data.

        # Partitions in main (left) memory are probed against partitions in secondary (right) memory.
        for i in range(self.left_table.size):
            partition = self.left_table.partitions[i]
            for record in partition.records:
                self.probeMainSecondaryMem(record, self.fileDescriptor_right, i, self.timestamp)

        # Partitions in main (right) memory are probed against partitions in secondary (left) memory.
        for i in range(self.right_table.size):
            partition = self.right_table.partitions[i]
            for record in partition.records:
                self.probeMainSecondaryMem(record, self.fileDescriptor_left, i, self.timestamp)

        # Partitions in secondary memory are probed.
        if (len(self.fileDescriptor_left) <= len(self.fileDescriptor_right)):
            self.probeSecondarySecondaryMem(self.fileDescriptor_left, self.fileDescriptor_right)
        else:
            self.probeSecondarySecondaryMem(self.fileDescriptor_right, self.fileDescriptor_left)

        # Delete files from secondary memory.
        for i in self.fileDescriptor_left:
            remove(self.fileDescriptor_left[i].file.name)

        for i in self.fileDescriptor_right:
            remove(self.fileDescriptor_right[i].file.name)

        # Put EOF in queue and exit.
        self.qresults.put("EOF")

    def probe(self, record, partition, var):
        # Probe a tuple if the partition is not empty.
        if partition:
            # For every record in the partition, check if it is duplicated.
            # Then, check if the tuple matches for every join variable.
            # If there is a join, concatenate the tuples and produce result.
            for r in partition.records:
#                if self.isDuplicated(record, r):
#                    break

                for v in var:
                    join = True
                    if record.tuple[v] != r.tuple[v]:
                        join = False
                        break

                if join:
                    res = record.tuple.copy()
                    res.update(r.tuple)
                    self.qresults.put(res)

    def probeMainSecondaryMem(self, record, filedescriptor2, partition, timestage2):
        # Probe a record against its corresponding partition in secondary memory.
        file2 = open(filedescriptor2[partition].file.name, 'r')
        records2 = file2.readlines()
        timestamps = filedescriptor2[partition].timestamps
        dtsLast = float("inf")

        for record2 in records2:
            (tuple2, ats2, dts2) = record2.split('|')
            tuple2 = eval(tuple2)
            ats2 = int(ats2)
            dts2 = int(dts2)
            dtsLast = dts2
            probedStage1 = False
            probedStage2 = False

            for v in self.vars:
                join = True
                if record.tuple[v] != tuple2[v]:
                    join = False
                    break

            if (join):
                # Check if the records were probed in stage 2.
                for pairTS in timestamps:
                    (dtsLast, probeTS) = pairTS
                    #if (isOverlapped((record.ats, record.dts), pairTS) and (record.dts >= pairTS[0])):
                    #if ((record.dts <= probeTS) and (dts2 > probeTS) and (ats2 <= probeTS)):
                    if ((dts2 <= probeTS) and (record.dts > probeTS) and (record.ats <= probeTS)):
                        probedStage2 = True
                        #print "Probed in stage 2:", (record.ats, record.dts), pairTS
                        break

                # Check if the records were probed in Stage 1.
                if (isOverlapped((record.ats, record.dts), (ats2, dts2))):
                    probedStage1 = True
                    #print "Probed in stage1:", (record.ats, record.dts), (ats2, dts2)

                # Produce result if records have not been produced.
                if (not(probedStage1) and not(probedStage2)):
                    res = record.tuple.copy()
                    res.update(tuple2)
                    self.qresults.put(res)
                    #if (len(res)==5): print "Produced at:", timestage2, res, (record.ats, record.dts), timestamps, (ats2,dts2)

        file2.close()
        return dtsLast

    def probeSecondarySecondaryMem(self, filedescriptor1, filedescriptor2):
        # Probe partitions in secondary memory.
        for partition in filedescriptor1.keys():
            if (filedescriptor2.has_key(partition)):
                file1 = open(filedescriptor1[partition].file.name, 'r')
                records1 = file1.readlines()
                timestamps1 = filedescriptor1[partition].timestamps

                for record1 in records1:
                    (tuple1, ats1, dts1) = record1.split('|')
                    tuple1 = eval(tuple1)
                    ats1 = int(ats1)
                    dts1 = int(dts1)
                    file2 = open(filedescriptor2[partition].file.name, 'r')
                    records2 = file2.readlines()
                    timestamps2 = filedescriptor2[partition].timestamps

                    for record2 in records2:
                        (tuple2, ats2, dts2) = record2.split('|')
                        tuple2 = eval(tuple2)
                        ats2 = int(ats2)
                        dts2 = int(dts2)

                        probedStage1 = False
                        probedStage2 = False

                        for v in self.vars:
                            join = True
                            if tuple1[v] != tuple2[v]:
                                join = False
                                break

                        if (join):
                            # Check if the records were probed in stage 2.
                            for pairTS in timestamps1:
                                (dtsLast, probeTS) = pairTS
                                # (dts1 <= dtsLast) and
                                if ((dts1 <= probeTS) and (dts2 > probeTS) and (ats2 <= probeTS)):
                                    probedStage2 = True
                                    break

                            for pairTS in timestamps2:
                                (dtsLast, probeTS) = pairTS
                                #(dts2 <= dtsLast) and
                                if ((dts2 <= probeTS) and (dts1 > probeTS) and (ats1 <= probeTS)):
                                    probedStage2 = True
                                    break

                            # Check if the records were probed in Stage 1.
                            if (isOverlapped((ats1, dts1), (ats2, dts2))):
                                probedStage1 = True

                            # Produce result if records have not been produced.
                            if (not(probedStage1) and not(probedStage2)):
                                res = tuple1.copy()
                                res.update(tuple2)
                                self.qresults.put(res)
                                #if (len(res)==5): print "Produced Sec Sec!:", res, (ats1, dts1), (ats2,dts2), timestamps1, timestamps2

                    file2.close()

                file1.close()

    def flushPartition(self):
        # Flush a Partition to secondary memory.

        # Choose a random victim from main memory, which size is not empty.
        victim_len = 0
        while (victim_len == 0):
            victim = randint(0, (self.left_table.size + self.right_table.size)-1)
            if (victim < self.left_table.size):
                victim_len = len(self.left_table.partitions[victim].records)
            else:
                victim_len = len(self.right_table.partitions[victim % self.right_table.size].records)

        if (victim < self.left_table.size):
            file_descriptor = self.fileDescriptor_left
            partition_to_flush = self.left_table.partitions[victim]
            self.memory_left = self.memory_left - len(partition_to_flush.records)
            self.left_table.partitions[victim] = Partition()

        else:
            victim = victim % self.right_table.size
            file_descriptor = self.fileDescriptor_right
            partition_to_flush = self.right_table.partitions[victim]
            self.memory_right = self.memory_right - len(partition_to_flush.records)
            self.right_table.partitions[victim] = Partition()

        # Update file descriptor
        if (file_descriptor.has_key(victim)):
            lenfile = file_descriptor[victim].size
            file = open(file_descriptor[victim].file.name, 'a')
            file_descriptor.update({victim: FileDescriptor(file, len(partition_to_flush.records) + lenfile, file_descriptor[victim].timestamps)})
        else:
            file = NamedTemporaryFile(suffix=".ht", prefix="", delete=False)
            file_descriptor.update({victim: FileDescriptor(file, len(partition_to_flush.records), set())})

        # Flush partition in file.
        for record in partition_to_flush.records:
            self.timestamp += 1
            tuple = str(record.tuple)
            ats   = str(record.ats)
            dts   = str(self.timestamp)
            #ats   = "%.40r" % (record.ats)
            #dts   = "%.40r" % (time())

            file.write(tuple + '|')
            file.write(ats + '|')
            file.write(dts + '\n')

        file.close()



"""
Created on Jul 10, 2011

Implements a Union operator.
The intermediate results are represented as lists.

@author: Maribel Acosta Deibe
"""

import itertools
from DeTrusty.Operators.Union import _Union
from multiprocessing import Queue


class Union(_Union):

    def __init__(self, vars_left, vars_right, distinct):
        self.left = []
        self.right = []
        self.results = []
        self.vars_left = set(vars_left)
        self.vars_right = set(vars_right)
        self.distinct = distinct

    def instantiate(self, d):
        newvars_left = self.vars_left - set(d.keys())
        newvars_right = self.vars_right - set(d.keys())
        return Union(newvars_left, newvars_right, self.distinct)

    def execute(self, qleft, qright, out, processqueue=Queue()):
        # Executes the Union operator.
        self.left = []
        self.right = []
        self.qresults = out

        # Initialize tuples.
        tuple1 = None
        tuple2 = None

        # Get the tuples from the queues.
        while (not(tuple1 == "EOF") or not(tuple2 == "EOF")):
            # Try to get tuple from left queue.
            if not(tuple1 == "EOF"):
                try:
                    tuple1 = qleft.get(False)
                    if not(tuple1 == "EOF"):
                        self.left.append(tuple1)
                        #print tuple1
                except Exception:
                    # This catch:
                    # Empty: in tuple2 = self.left.get(False), when the queue is empty.
                    pass

            # Try to get tuple from right queue.
            if not(tuple2 == "EOF"):
                try:
                    tuple2 = qright.get(False)
                    if not(tuple2 == "EOF"):
                        self.right.append(tuple2)
                        #print tuple2
                except Exception:
                    # This catch:
                    # Empty: in tuple2 = self.right.get(False), when the queue is empty.
                    pass

        if (self.vars_left == self.vars_right):
            self.sameVariables()
        else:
            self.differentVariables()
        #print "cardinalidad: "+str(len(self.results))
        # Put all the results in the output queue.
        while self.results:
            self.qresults.put(self.results.pop(0))
        #print "cardinalidad: "+str(len(self.results))

        # Put EOF in queue and exit.
        self.qresults.put("EOF")

    def sameVariables(self):
        # Executes the Union operator when the variables are the same.

        # Concatenates and sort left and right lists.
        if self.distinct:
            for t in self.left:
                if not (t in self.right):
                    self.results.append(t)
            self.results.extend(self.right)
        else:
            self.results = self.left + self.right

    def differentVariables(self):
        # Executes the Union operator when the variables are not the same.

        # Iterates elements of left list.
        for tuple1 in self.left:
            res = {}
            for v in self.vars_right:
                res.update({v:''})
            res.update(tuple1)
            self.results.append(res)

        # Iterates elements of right list.
        for tuple2 in self.right:
            res = {}
            for v in self.vars_left:
                res.update({v:''})
            res.update(tuple2)
            self.results.append(res)



"""
NestedHashJoin.py

Implements a depending operator, similar to block nested join and symmetric
hash join

Autor: Gabriela Montoya
Date: July 18th, 2012
"""

from multiprocessing import Queue, Process
from time import time
import string, sys
from queue import Empty
from DeTrusty.Operators.Join import Join
from .OperatorStructures import Table, Partition, Record


class NestedHashJoin(Join):

    def __init__(self, vars):
        self.left_table = dict()
        self.right_table = dict()
        self.qresults = Queue()
        self.vars = vars

    def instantiate(self, d):
        newvars = self.vars - set(d.keys())
        return NestedHashJoin(newvars)

    def execute(self, left_queue, right_operator, out, processqueue=Queue()):
        self.left_queue = left_queue
        self.right_operator = right_operator
        self.qresults = out
        # print "right_operator", right_operator
        tuple1 = None
        tuple2 = None
        right_queues = dict()

        while (not (tuple1 == "EOF") or (len(right_queues) > 0)):
            # Try to get and process tuple from left queue
            if not (tuple1 == "EOF"):
                try:
                    tuple1 = self.left_queue.get(False)
                    # print "tuple1: "+str(tuple1)
                    instance = self.probeAndInsert1(tuple1, self.right_table, self.left_table, time())
                    if instance:  # the join variables have not been used to
                        # instantiate the right_operator

                        new_right_operator = self.makeInstantiation(tuple1, self.right_operator)
                        # print "new op: "+str(new_right_operator)
                        resource = self.getResource(tuple1)
                        queue = Queue()
                        right_queues[resource] = queue
                        # print "new_right_operator.__class__", new_right_operator.__class__
                        # print "new_right_operator.left.__class__", new_right_operator.left.__class__
                        new_right_operator.execute(queue)
                        # p2 = Process(target=new_right_operator.execute, args=(queue,))
                        # p2.start()
                except Empty:
                    pass
                except TypeError:
                    # TypeError: in resource = resource + tuple[var], when the tuple is "EOF".
                    pass
                except Exception as e:
                    # print "Unexpected error:", sys.exc_info()[0]
                    print(e)
                    pass

            toRemove = []  # stores the queues that have already received all its tuples

            for r in right_queues:
                try:
                    q = right_queues[r]
                    tuple2 = q.get(False)
                    # print "tuple2", tuple2
                    if (tuple2 == "EOF"):
                        toRemove.append(r)
                    else:
                        self.probeAndInsert2(r, tuple2, self.left_table,
                                             self.right_table, time())
                except Exception:
                    # This catch:
                    # Empty: in tuple2 = self.right.get(False), when the queue is empty.
                    # TypeError: in att = att + tuple[var], when the tuple is "EOF".
                    # print "Unexpected error:", sys.exc_info()[0]
                    pass

            for r in toRemove:
                del right_queues[r]
        # Put EOF in queue and exit.
        self.qresults.put("EOF")
        return

    def getResource(self, tuple):
        resource = ''
        for var in self.vars:
            resource = resource + tuple[var]
        return resource

    def makeInstantiation(self, tuple, operator):
        d = {}
        for var in self.vars:
            v = tuple[var]
            if v.find("http") == 0:  # uris must be passed between < .. >
                v = "<" + v + ">"
            else:
                v = '"' + v + '"'
            d[var] = v
        new_operator = operator.instantiate(d)
        return new_operator

    def probeAndInsert1(self, tuple, table1, table2, time):
        record = Record(tuple, time, 0)
        r = self.getResource(tuple)
        if r in table1:
            records = table1[r]
            for t in records:
                if t.ats > record.ats:
                    continue
                x = t.tuple.copy()
                x.update(tuple)
                self.qresults.put(x)
        p = table2.get(r, [])
        i = (p == [])
        p.append(record)
        table2[r] = p
        return i

    def probeAndInsert2(self, resource, tuple, table1, table2, time):
        record = Record(tuple, time, 0)
        if resource in table1:
            records = table1[resource]
            for t in records:
                if t.ats > record.ats:
                    continue
                x = t.tuple.copy()
                x.update(tuple)
                self.qresults.put(x)
        p = table2.get(resource, [])
        p.append(record)
        table2[resource] = p



"""
Created on Jul 10, 2011

Implements the Xnjoin operator.
The intermediate results are represented in a queue.

@author: Maribel Acosta Deibe
"""

from multiprocessing import Queue
from time import time
from DeTrusty.Operators.Join import Join
from .OperatorStructures import Record, RJTTail


class Xnjoin(Join):

    def __init__(self, vars):
        self.left_table  = dict()
        self.right_table = dict()
        self.qresults    = Queue()
        self.vars        = vars

    def instantiate(self, d):
        newvars = self.vars - set(d.keys())
        return Xnjoin(newvars)

    def instantiateFilter(self, instantiated_vars, filter_str):
        newvars = self.vars - set(instantiated_vars)
        return Xnjoin(newvars)

    def execute(self, left, right, out, processqueue=Queue()):
        # Executes the Xgjoin.
        self.left     = left
        self.right    = right
        self.qresults = out

        # Get tuples from queue.
        tuple = self.left.get(True)

        # Get the tuples from the queues.
        while (not(tuple == "EOF")):
            self.stage1(tuple, self.left_table, self.right_table)
            tuple = self.left.get(True)

        # Perform the last probes.
        self.stage3()


    def stage1(self, tuple, tuple_rjttable, other_rjttable):
        # Stage 1: While one of the sources is sending data.

        # Get the resource associated to the tuples.
        resource = ''
        for var in self.vars:
            resource = resource + tuple[var]

        # Probe the tuple against its RJT table.
        probeTS = self.probe(tuple, resource, tuple_rjttable, other_rjttable)

        # Create the records.
        record = Record(tuple, probeTS, time())

        # Insert the record in the other RJT table.
        # TODO: use RJTTail. Check ProbeTS
        if resource in other_rjttable:
            other_rjttable.get(resource).updateRecords(record)
            other_rjttable.get(resource).setRJTProbeTS(probeTS)
            #other_rjttable.get(resource).append(record)
        else:
            tail = RJTTail(record, float("inf"))
            other_rjttable[resource] = tail
            #other_rjttable[resource] = [record]

    def stage2(self):
        # Stage 2: When both sources become blocked.
        pass

    def stage3(self):
        # Stage 3: When both sources sent all the data.

        # Put EOF in queue and exit.
        self.qresults.put("EOF")
        return

    def probe(self, tuple, resource, rjttable, other_rjttable):
        probeTS = time()

        # If the resource is in table, produce results.
        if resource in rjttable:
            rjttable.get(resource).setRJTProbeTS(probeTS)
            list_records = rjttable[resource].records
            #list_records = rjttable[resource]

            for record in list_records:
                res = record.tuple.copy()
                res.update(tuple)
                self.qresults.put(res)

        # If not, contact the source.
        else:
            instances = []
            for v in self.vars:
                instances = instances + [tuple[v]]

            # Contact the source.
            qright = Queue()
            # print "instances: "+str(instances)
            self.right.execute(self.vars, instances, qright)

            # Get the tuples from right queue.
            rtuple = qright.get(True)
            while (not(rtuple == "EOF")):
                # Build answer and produce it.
                rtuple_copy = rtuple.copy()
                rtuple_copy.update(tuple)
                self.qresults.put(rtuple_copy)

                # Create and insert the record in the left RJT table.
                record = Record(rtuple, probeTS, time())
                if resource in rjttable:
                    other_rjttable.get(resource).updateRecords(record)
                    other_rjttable.get(resource).setRJTProbeTS(probeTS)
                else:
                    tail = RJTTail(record, float("inf"))
                    other_rjttable[resource] = tail

                rtuple = qright.get(True)

        return probeTS

#    def fase1(self, tuple1, tuple2):
#
#        # Get the resource associated to the tuples
#        resource1 = ''
#        resource2 = ''
#        for var in self.vars:
#            resource1 = resource1 + tuple1[var]
#            resource2 = resource2 + tuple2[var]
#
#        # Probe the tuple against its RJT table.
#        # Create the records.
#        # Insert the records in RJT tables.
#        probeTS1 = self.probe(tuple1, resource1, self.tablaizq)
#        record1 = Record(tuple1, probeTS1, time.time())
#
#        if resource1 in self.tablader:
#            self.tablader.get(resource1).append(record1)
#        else:
#            self.tablader[resource1] = [record1]
#
#        # Probe the tuple against its RJT table.
#        # Create the records.
#        # Insert the records in RJT tables.
#
#        probeTS2 = self.probe(tuple2, resource2, self.tablader)
#        record2 = Record(tuple2, probeTS2, time.time())
#        if resource2 in self.tablaizq:
#            self.tablaizq.get(resource2).append(record2)
#        else:
#            self.tablaizq[resource2] = [record2]
#        #other_rjttable.insertRecord(resource, record)



"""
Created on Jul 10, 2011

Implements the structures used by the Non-Blocking Operators.

@author: Maribel Acosta Deibe
"""

from multiprocessing import Manager


class Record(object):
    """
    Represents a structure that is inserted into the hash table.
    It is composed by a tuple, ats (arrival timestamp) and
    dts (departure timestamp).
    """

    def __init__(self, tuple, ats, dts):
        self.tuple = tuple
        self.ats = ats
        self.dts = dts

    def __repr__(self):
        return "(" + str(self.tuple) + ", " + str(self.ats) + ", " + str(self.dts) + ")"


class Partition(object):
    """
    Represents a bucket of the hash table.
    It is composed by a list of records, and a list of timestamps
    of the form {DTSlast, ProbeTS}
    """

    def __init__(self):
        self.records = []  # List of records
        self.timestamps = []  # List of the form {DTSlas, ProbeTS}


class Table(object):
    """
    Represents a hash table.
    It is composed by a list of partitions (buckets) of size n,
    where n is specified in "size".
    """

    def __init__(self):
        self.size = 1
        self.partitions = [Partition() for x in range(self.size)]

    def getSize(self):
        return self.size

    def insertRecord(self, i, value):
        self.partitions[i].records.append(value)


class FileDescriptor(object):
    """
    Represents the description of a file, that contains a RJT in sec mem.
    It is composed by the name of the file, the associated resource,
    the current size (number of tuples), and the timestamp of the last
    RJTTail that have been flushed.
    """

    def __init__(self, file, size, timestamps):
        # self.manager = Manager()
        self.file = file
        self.size = size
        self.timestamps = timestamps
        # self.timestamps = self.manager.list() #set()  #[]

    def getSize(self):
        return self.size


# class PairTS(object):
#    '''
#    Represents the description pair of {DTSlast, probeTS} to detect tuples
#    that could be generated in a second stage. DTSlast is the D T  S value of
#    the last tuple of the disk-resident portion that was used to probe the
#    memory-resident tuples, and P robeT S is the timestamp value at the time
#    that the second stage was executed,
#    '''
#    def __init__(self, dtsLast, probeTS):
#        self.dtsLast = dtsLast
#        self.probeTS = probeTS


def isOverlapped(X, Y):
    (x1, x2), (y1, y2) = X, Y
    return (partiallyOverlapped((x1, x2), (y1, y2)) or
            fullyOverlapped((x1, x2), (y1, y2)))


def partiallyOverlapped(X, Y):
    (x1, x2), (y1, y2) = X, Y
    return ((x1 <= y1 <= x2 <= y2) or (y1 <= x1 <= y2 <= x2))


def fullyOverlapped(X, Y):
    (x1, x2), (y1, y2) = X, Y
    return ((x1 <= y1 <= y2 <= x2) or (y1 <= x1 <= x2 <= y2))



"""
Created on Mar 03, 2014

Implements the Xorderby operator.
The intermediate results are represented in a queue.

@author: Maribel Acosta Deibe
"""

from multiprocessing import Queue
import datetime

data_types = {
    'integer': (int, 'numerical'),
    'decimal': (float, 'numerical'),
    'float': (float, 'numerical'),
    'double': (float, 'numerical'),
    'string': (str, str),
    'boolean': (bool, bool),
    'dateTime': (datetime, datetime),
    'nonPositiveInteger': (int, 'numerical'),
    'negativeInteger': (int, 'numerical'),
    'long': (int, 'numerical'),
    'int': (int, 'numerical'),
    'short': (int, 'numerical'),
    'byte': (bytes, bytes),
    'nonNegativeInteger': (int, 'numerical'),
    'unsignedLong': (int, 'numerical'),
    'unsignedInt': (int, 'numerical'),
    'unsignedShort': (int, 'numerical'),
    'unsignedByte': (bytes, bytes),  # TODO: this is not correct
    'positiveInteger': (int, 'numerical')
}


class Xorderby(object):

    def __init__(self, args):
        self.input = Queue()
        self.qresults = Queue()
        self.args = args  # List of type Argument.
        # print "self.args", self.args

    def execute(self, left, dummy, out, processqueue=Queue()):
        # Executes the Xorderby.
        self.left = left
        self.qresults = out
        results = []
        results_copy = []

        # Read all the results.
        tuple = self.left.get(True)
        # print "tuple", tuple
        tuple_id = 0
        while (tuple != "EOF"):
            results_copy.append(tuple)
            res = {}
            res.update(tuple)
            # print "tuple", tuple
            for arg in self.args:
                res.update({arg.name[1:]: self.extractValue(tuple[arg.name[1:]])})
            res.update({'__id__': tuple_id})
            results.append(res)
            tuple_id = tuple_id + 1
            tuple = self.left.get(True)

        # Sorting.
        self.args.reverse()
        # print "en order by ",self.args
        for arg in self.args:
            order_by = "lambda d: (d['" + arg.name[1:] + "'])"
            results = sorted(results, key=eval(order_by), reverse=arg.desc)

        # Add results to output queue.
        for tuple in results:
            self.qresults.put(results_copy[tuple['__id__']])

        # Put EOF in queue and exit.
        self.qresults.put("EOF")
        return

    def extractValue(self, val):
        pos = val.find("^^")

        # Handles when the literal is typed.
        if (pos > -1):
            for t in data_types.keys():
                if (t in val[pos]):
                    (python_type, general_type) = data_types[t]

                    if (general_type == bool):
                        return val[:pos]

                    else:
                        return python_type(val[:pos])

        # Handles non-typed literals.
        else:
            return str(val)
